/* pam_handlers.c -- pam config file parsing and module loading */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dlfcn.h>

#include "pam_appl.h"
#include "pam_private.h"

#define BUF_SIZE      1024
#define MODULE_CHUNK  4

#ifdef DEBUG
#define D(x) _pam_debug x
#else
#define D(x)
#endif

int _pam_assemble_line(FILE *f, char *buf, int buf_len);

void _pam_free_handlers_aux(struct handler **hp);

int _pam_add_handler(pam_handle_t *pamh, int type,
		     int control_flag, char *mod_path,
		     int argc, char **argv, int argvlen);

/* Values for type */
#define PAM_AUTH    1
#define PAM_SESS    2
#define PAM_ACCT    4
#define PAM_PASS    8

/* Parse config file, allocate handler structures, dlopen() */
int _pam_init_handlers(pam_handle_t *pamh)
{
    FILE *f;
    char buf[BUF_SIZE];
    char *tok, *mod_path;
    int x, module_type, control_flag;
    int other, modules_found;
    char *service_name = pamh->service_name;
    int res;
#ifdef DEBUG
    int y;
#endif

    int argc;
    char **argv;
    int argvlen;

    /* Return immediately if everything is already loaded */
    if (pamh->handlers.handlers_loaded) {
	return PAM_SUCCESS;
    }

    D(("_pam_init_handlers: initializing"));
    
    /* First clean the service structure */

    _pam_free_handlers(pamh);
    if (! pamh->handlers.module) {
	pamh->handlers.module = malloc(MODULE_CHUNK *
				       sizeof(struct loaded_module));
	pamh->handlers.modules_allocated = MODULE_CHUNK;
	pamh->handlers.modules_used = 0;
    }
    
    /* Now parse the config file and add handlers */

    modules_found = 0;
    f = fopen(PAM_CONFIG, "r");
    while ((x = _pam_assemble_line(f, buf, BUF_SIZE)) > 0) {
	D(("_pam_init_handler: LINE: %s", buf));
	tok = strtok(buf, " \n\t");
	other = !strcmp(tok, "OTHER");
	if (!strcmp(tok, service_name) || other) {
	    /* This is the service we are looking for */
	    D(("_pam_init_handlers: Found pam.conf entry for: %s", tok));

	    tok = strtok(NULL, " \n\t");
	    if (!strcmp("auth", tok)) {
		module_type = PAM_AUTH;
	    } else if (!strcmp("session", tok)) {
		module_type = PAM_SESS;
	    } else if (!strcmp("account", tok)) {
		module_type = PAM_ACCT;
	    } else if (!strcmp("password", tok)) {
		module_type = PAM_PASS;
	    } else {
		/* Illegal module type */
		D(("_pam_init_handlers: bad module type: %s", tok));
		_pam_log_error("illegal module type: %s", tok);
		return(PAM_ABORT);
	    }
	    D(("Using pam.conf entry: %s", tok));

	    /* Only accept OTHER if we do not yet have this module_type */
	    if (other && (modules_found & module_type)) {
		/* ignore */
		D(("_pam_init_handlers: (OTHER) skipping entry: %s", tok));
	    } else {
		modules_found |= module_type;
	    
		tok = strtok(NULL, " \n\t");
		if (!strcmp("required", tok)) {
		    control_flag = PAM_REQUIRED;
		} else if (!strcmp("optional", tok)) {
		    control_flag = PAM_OPTIONAL;
		} else if (!strcmp("sufficient", tok)) {
		    control_flag = PAM_SUFFICIENT;
		} else {
		    /* Illegal flag */
		    D(("_pam_init_handlers: bad control flag: %s", tok));
		    _pam_log_error("illegal control flag: %s", tok);
		    return(PAM_ABORT);
		}
		
		tok = strtok(NULL, " \n\t");
		mod_path = tok;
		
		tok += strlen(tok) + 1;
		argvlen = pam_mkargv(tok, &argv, &argc);

#ifdef DEBUG
		D(("CONF: %s %d %d %s %d", service_name, module_type,
		       control_flag, mod_path, argc));
		for (y = 0; y < argc; y++) {
		    D(("CONF: %s", argv[y]));
		}
#endif

		res = _pam_add_handler(pamh, module_type, control_flag,
				       mod_path, argc, argv, argvlen);
		if (res != PAM_SUCCESS) {
		    _pam_log_error("error loading %s", mod_path);
		    return res;
		}
	    }
	}
    }
    fclose(f);

    if (x < 0) {
	/* Read error */
	_pam_log_error("error reading pam.conf");
	return PAM_ABORT;
    }

    if (modules_found != (PAM_AUTH | PAM_SESS | PAM_ACCT | PAM_PASS)) {
	D(("modules_found = %d", modules_found));
	_pam_log_error("incomplete pam.conf spec for: %s", service_name);
	return PAM_ABORT;
    }
    
    pamh->handlers.handlers_loaded = 1;
    return PAM_SUCCESS;
}

int _pam_assemble_line(FILE *f, char *buffer, int buf_len)
{
    char *p = buffer;
    char *s, *os;
    int used = 0;

    while (1) {
	if (used >= buf_len) {
	    /* Overflow */
	    D(("_pam_assemble_line: overflow"));
	    return -1;
	}
	if (!fgets(p, buf_len - used, f)) {
	    if (used) {
		/* Incomplete read */
		return -1;
	    } else {
		/* EOF */
		return 0;
	    }
	}
	s = p + strspn(p, " \n\t");
	if (*s && (*s != '#')) {
	    /* Check for backslash */
	    os = s;
	    s += strlen(s);
	    while ((*s == ' ') || (*s == '\t')) {
		s--;
	    }
	    if (*s == '\\') {
		*s++ = ' ';
		*s = '\0';
		used += strlen(os);
		p = s;
	    } else {
		/* End of the line! */
		used += strlen(os);
		break;
	    }
	} else {
	    /* Nothing in this line */
	    /* Don't move p         */
	}
    }

    return used;
}

/* XXX - _pam_add_handler() is where we would handle the static libs */

int _pam_add_handler(pam_handle_t *pamh, int type,
		     int control_flag, char *mod_path,
		     int argc, char **argv, int argvlen)
{
    struct loaded_module *mod = pamh->handlers.module;
    int x = 0;
    struct handler **handler_p;
    struct handler **handler_p2;
    char *sym, *sym2;
    int (*func)(pam_handle_t *, int, int, char **);
    int (*func2)(pam_handle_t *, int, int, char **);
    
    /* First, ensure the module is loaded */
    while (x < pamh->handlers.modules_used) {
	if (!strcmp(mod[x].name, mod_path)) {
	    break;
	}
	x++;
    }
    if (x == pamh->handlers.modules_used) {
	/* Not found */
	if (pamh->handlers.modules_allocated == pamh->handlers.modules_used) {
	    pamh->handlers.modules_allocated += MODULE_CHUNK;
	    pamh->handlers.module = realloc(pamh->handlers.module,
					    pamh->handlers.modules_allocated);
	}
	mod = &(pamh->handlers.module[x]);
	D(("_pam_add_handler: dlopen(%s)", mod_path));
	mod->dl_handle = dlopen(mod_path, RTLD_NOW);
	if (! mod->dl_handle) {
	    D(("_pam_add_handler: dlopen(%s) failed", mod_path));
	    _pam_log_error("unable to dlopen(%s)", mod_path);
	    return(PAM_ABORT);
	} else {
	    mod->name = strdup(mod_path);
	    pamh->handlers.modules_used++;
	}
    }

    /* mod points to the struct loaded_module we care about */
    /* Now add the handler based on mod->dlhandle and type  */

    handler_p2 = NULL;
    func2 = NULL;
    sym2 = NULL;
    switch (type) {
    case PAM_AUTH:
	handler_p = &pamh->handlers.authenticate;
	sym = "pam_sm_authenticate";
	handler_p2 = &pamh->handlers.setcred;
	sym2 = "pam_sm_setcred";
	break;
    case PAM_SESS:
	handler_p = &pamh->handlers.open_session;
	sym = "pam_sm_open_session";
	handler_p2 = &pamh->handlers.close_session;
	sym2 = "pam_sm_close_session";
	break;
    case PAM_ACCT:
	handler_p = &pamh->handlers.acct_mgmt;
	sym = "pam_sm_acct_mgmt";
	break;
    case PAM_PASS:
	handler_p = &pamh->handlers.chauthtok;
	sym = "pam_sm_chauthtok";
	break;
    default:
	/* Illegal module type */
	D(("_pam_add_handler: illegal module type %d", type));
	return PAM_ABORT;
    }

    if (! (func = dlsym(mod->dl_handle, sym))) {
	_pam_log_error("unable to resolve symbol: %s", sym);
	return(PAM_ABORT);
    }
    if (sym2) {
	if (! (func2 = dlsym(mod->dl_handle, sym2))) {
	    _pam_log_error("unable to resolve symbol: %s", sym2);
	    return(PAM_ABORT);
	}
    }
    
    while (*handler_p) {
	handler_p = &((*handler_p)->next);
    }
    *handler_p = malloc(sizeof(struct handler));
    (*handler_p)->func = func;
    (*handler_p)->control_flag = control_flag;
    (*handler_p)->argc = argc;
    (*handler_p)->argv = argv;
    (*handler_p)->next = NULL;

    if (handler_p2) {
	while (*handler_p2) {
	    handler_p2 = &((*handler_p2)->next);
	}
	*handler_p2 = malloc(sizeof(struct handler));
	(*handler_p2)->func = func2;
	(*handler_p2)->control_flag = control_flag;
	(*handler_p2)->argc = argc;
	if (argv) {
	    (*handler_p2)->argv = malloc(argvlen);
	    memcpy((*handler_p2)->argv, argv, argvlen);
	} else {
	    (*handler_p2)->argv = NULL;
	}
	(*handler_p2)->next = NULL;
    }
    
    return PAM_SUCCESS;
}

/* Free various allocated structures and dlclose() the libs */
int _pam_free_handlers(pam_handle_t *pamh)
{
    struct loaded_module *mod = pamh->handlers.module;

    /* Close all loaded modules */
    while (pamh->handlers.modules_used) {
	D(("_pam_free_handlers: dlclose(%s)", mod->name));
	free(mod->name);
	dlclose(mod->dl_handle);
	mod++;
	pamh->handlers.modules_used--;
    }

    /* Free all the handlers */
    _pam_free_handlers_aux(&(pamh->handlers.authenticate));
    _pam_free_handlers_aux(&(pamh->handlers.setcred));
    _pam_free_handlers_aux(&(pamh->handlers.acct_mgmt));
    _pam_free_handlers_aux(&(pamh->handlers.open_session));
    _pam_free_handlers_aux(&(pamh->handlers.close_session));
    _pam_free_handlers_aux(&(pamh->handlers.chauthtok));

    /* Indicate that handlers are not initialized for this pamh */
    pamh->handlers.handlers_loaded = 0;

    return PAM_SUCCESS;
}

void _pam_start_handlers(pam_handle_t *pamh)
{
    /* Indicate that handlers are not initialized for this pamh */
    pamh->handlers.handlers_loaded = 0;

    pamh->handlers.modules_allocated = 0;
    pamh->handlers.modules_used = 0;
    pamh->handlers.module = NULL;

    pamh->handlers.authenticate = NULL;
    pamh->handlers.setcred = NULL;
    pamh->handlers.acct_mgmt = NULL;
    pamh->handlers.open_session = NULL;
    pamh->handlers.close_session = NULL;
    pamh->handlers.chauthtok = NULL;
}

void _pam_free_handlers_aux(struct handler **hp)
{
    struct handler *h = *hp;
    struct handler *last;

    while (h) {
	last = h;
	free(h->argv);  /* This is all alocated in a single chunk */
	h = h->next;
	free(last);
    }

    *hp = NULL;
}
